/*
 * MySQLClient.c
 *
 *  Created on: 14 Μαρ 2012
 *      Author: nick
 */
#include <cr_section_macros.h>

#include "MySQLClient.h"
#include "lwip/api.h"
#include "lwip/tcpip.h"
#include "lwip/sockets.h"
#include "lwip/def.h"

#include "sha1.h"
#include "string.h"
#include "util/debug.h"


// http://forge.mysql.com/wiki/MySQL_Internals_ClientServer_Protocol#Handshake_Initialization_Packet
#define MYSQL_TIMEOUT_MS 10000
#define MYSQL_PORT 3306

#define BUF_SIZE 128

#define CLIENT_LONG_PASSWORD     	1
#define CLIENT_CONNECT_WITH_DB   	8
#define CLIENT_PROTOCOL_41       	512
#define CLIENT_INTERACTIVE         	1024
#define CLIENT_SECURE_CONNECTION 	32768
#define CLIENT_RESERVED         	16384

#define MIN(a,b) ((a)<(b)?(a):(b))
#define ABS(a) (((a)>0)?(a):0)

/*
Bytes                        Name
 -----                        ----
 1                            protocol_version
 n (Null-Terminated String)   server_version
 4                            thread_id
 8                            scramble_buff
 1                            (filler) always 0x00
 2                            server_capabilities
 1                            server_language
 2                            server_status
 2                            server capabilities (two upper bytes)
 1                            length of the scramble
10                            (filler)  always 0
 n                            rest of the plugin provided data (at least 12 bytes)
 1                            \0 byte, terminating the second part of a scramble
 */


//MySQL commands
#define COM_QUIT  0x01 //Exit
#define COM_QUERY 0x03 //Execute an SQL query

static err_t handleHandshake(struct netconn *conn);
static err_t sendAuth(struct netconn *conn, const char *user, const char *password,const char *db);
static enum MySQLResult handleAuthResult(struct netconn *conn);
static struct netbuf *readSock(struct netconn *conn,byte *buf, unsigned short *buflen);
err_t writeSock(struct netconn *conn);
static err_t writeSockCmd(struct netconn *conn, byte command, char *arg);
static void mysql_execute(void *arg);

__BSS(RAM2) static byte m_buf[BUF_SIZE];
static byte *m_pPos;
static uint16_t m_len;
static uint16_t m_packetId;

struct mysql_cmd{

	struct netconn *conn;
	byte command;
	char *bytes;
	char *arg;
	int len;
	char *msg;
	sys_sem_t lock;
	err_t err;
};

struct netconn *MySqlOpen(struct netif *net_eth,char *host, const char *user, const char *password, const char *db, void (*pMethod)(enum MySQLResult)){


	struct netconn *conn;
	ip_addr_t remoteHost;
	err_t err;

	remoteHost.addr = inet_addr(host);

	memset(m_buf,0,BUF_SIZE);
	m_len = 0;
	m_pPos = m_buf;
	m_packetId = 0;

	DEBUGF(MYSQL_DEBUG,("Connecting to MySQL...\n"));

	conn = netconn_new(NETCONN_TCP);
	if(conn == NULL || net_eth == NULL){
		netconn_delete(conn);
		return NULL;
	}

	 /* nonblocking */
//	opt = O_NONBLOCK;
//	ret = lwip_fcntl(s, F_SETFL, opt);
	int rc1 = netconn_bind(conn,&net_eth->ip_addr,0);
	int rc2 = netconn_connect(conn,&remoteHost,MYSQL_PORT);

	if(rc1 != ERR_OK || rc2 != ERR_OK ){
		DEBUGF(MYSQL_DEBUG,("Error connecting to MySQL\n"));
		netconn_delete(conn);
		return  NULL;
	}

	/* set recv timeout (500 ms) */
//	opt = MYSQL_TIMEOUT_MS;
//	ret = lwip_setsockopt(s, SOL_SOCKET, SO_RCVTIMEO, &opt, sizeof(int));


	err = handleHandshake(conn);

	if(err == ERR_OK){
		if( sendAuth( conn, user, password, db ) == ERR_OK) {
			err  = handleAuthResult(conn);
			if(err == MYSQL_AUTHFAILED){
				netconn_delete(conn);
				conn = NULL;
			}
		}else{
			netconn_delete(conn);
			conn = NULL;
		}
	}else{
		netconn_delete(conn);
		conn = NULL;
	}

	return conn;

}


static err_t handleHandshake(struct netconn *conn){


	unsigned short len;

	struct netbuf *in  = readSock(conn,m_buf,&len);

	if( ! (( m_len > 1 ) && ( memchr( m_buf + 1, 0, m_len ) != NULL )) )
	{
		DEBUGF(MYSQL_DEBUG,("Connected but could not find pcsz...\n"));
		netbuf_delete( in );
		return ERR_CONN;
	}


	DEBUGF(MYSQL_DEBUG,("Connected to server: %d bytes read ; Protocol version %d, mysql: %s.\n", m_len, m_buf[0],m_buf[1]));
	m_pPos = (byte*) memchr( (char*)(m_buf + 1), 0, m_len ) + 1;
	netbuf_delete( in );

	return ERR_OK;


}

static err_t sendAuth(struct netconn *conn, const char *user, const char *password,const char *db){

	int i;
	err_t err;

	uint16_t serverFlags = *((uint16_t*)&m_pPos[13]);

	DEBUGF(MYSQL_DEBUG,("Server capabilities are %04X.\n", serverFlags));

	if( ! (serverFlags & (CLIENT_RESERVED | CLIENT_PROTOCOL_41	| CLIENT_SECURE_CONNECTION  )) )   //  New 4.1 protocol
	{
		//We only support protocol >= mysql-4.1
		DEBUGF(MYSQL_DEBUG,("We only support protocol >= mysql-4.1\n"));
		return ERR_CONN;
	}

	uint32_t clientFlags = CLIENT_CONNECT_WITH_DB | CLIENT_PROTOCOL_41 | CLIENT_SECURE_CONNECTION | CLIENT_INTERACTIVE;

	//if(serverFlags & CLIENT_LONG_PASSWORD)

	DEBUGF(MYSQL_DEBUG,("Using auth 4.1+\n"));
	//Encrypt pw using scramble
	byte scramble[20+20]={0};
	memcpy(scramble, m_pPos+4, 8);
	memcpy(scramble+8, m_pPos+31, 12); // *(m_pPos+43) == 0 (zero-terminated char*)

   byte stage1_hash[20] = {0};
   sha1( (byte*)password, strlen(password), stage1_hash );

   sha1( stage1_hash, 20, ((byte*)scramble + 20) );

   byte token[20] = {0};
   sha1( scramble, 40, (unsigned char*)token );

   for(i=0;i<20;i++)
     token[i] = token[i] ^ stage1_hash[i];

   clientFlags |= CLIENT_LONG_PASSWORD;

   DEBUGF(MYSQL_DEBUG,("Building response\n"));
   //Build response

   memset(m_buf,0,BUF_SIZE);
   //BE
   *((uint32_t*)&m_buf[0]) = clientFlags;
   *((uint32_t*)&m_buf[4]) = BUF_SIZE; //Max packets size
   m_buf[8] = 8; //latin1 charset
   memset((char*)(m_buf+9),0,23);
   strcpy((char*)(m_buf+32),user);
   m_pPos = m_buf + 32 + strlen(user) + 1;
   m_pPos[0] = 20;
   memcpy((char*)&m_pPos[1],token,20);
   strcpy((char*)(m_pPos+21),db);
   m_len = 32 + strlen(user) + 1 + 21 + strlen(db) + 1;

   //Save first part of scramble in case we need it again
   memcpy(&m_buf[BUF_SIZE-8], scramble, 8);

   err = writeSock(conn);

   return err;

}

static enum MySQLResult handleAuthResult(struct netconn *conn){

	unsigned short len;

	struct netbuf *in  = readSock(conn,m_buf,&len);

	if(len<2)
	{
		DEBUGF(MYSQL_DEBUG,("Response too short..\n"));
		netbuf_delete( in );
	    return MYSQL_PRTCL;

	}

	DEBUGF(MYSQL_DEBUG,("RC=%d ",m_buf[0]));

	if(m_buf[0]==0)
	{
	   	m_buf[m_len] = 0;
	    m_pPos = m_buf + 1;
	    m_pPos += m_buf[1] +1;
	    m_pPos += m_pPos[0];
	    m_pPos += 1;
	    DEBUGF(MYSQL_DEBUG,("(OK) : Server status %d, Message : %s\n", *((uint16_t*)&m_pPos[0]), m_pPos+4));
	    netbuf_delete( in );
	    return MYSQL_OK;
	 }
	 else{
	    m_buf[m_len] = 0;
	    DEBUGF(MYSQL_DEBUG,("(Error %d) : %s\n", *((uint16_t*)&m_buf[1]), &m_buf[9])); //LE
	    netbuf_delete( in );
	    return MYSQL_AUTHFAILED;

	 }

}

enum MySQLResult MySqlExecute(struct netconn *conn, byte command, char *arg, int len){


	struct mysql_cmd cmd;

	cmd.conn = conn;
	cmd.command = command;
	cmd.arg = arg;
	cmd.len = len;


//	sys_sem_new(&cmd.lock, 0);
//	sys_thread_t t = sys_thread_new("mysql_ex", mysql_execute, &cmd, 200, 0);
//	sys_sem_wait(&cmd.lock);

//	sys_arch_thread_remove(t);
	mysql_execute(&cmd);
//	sys_sem_free(&cmd.lock);

	return cmd.err;

}

static void mysql_execute(void *arg){

	struct mysql_cmd *cmd = (struct mysql_cmd*)arg;
	struct netbuf *in;
	unsigned short len;

	DEBUGF(MYSQL_DEBUG,("Sending command %d, payload of len %d\n", cmd->command, cmd->len));

//	for(;;){

		m_packetId=0;//Reset packet ID (New sequence)
		m_len = 1 + cmd->len;

		if(writeSockCmd(cmd->conn,cmd->command,cmd->arg) == ERR_OK){
			in = readSock(cmd->conn,m_buf,&len);
			if(in != NULL ){
				if(m_len<2)
				{
					DEBUGF(MYSQL_DEBUG,("Response too short..\n"));
					cmd->err = MYSQL_PRTCL;
					sys_sem_signal(&cmd->lock);
				}

				cmd->msg = (char*)&m_buf[9];
				DEBUGF(MYSQL_DEBUG,("RC=%d ",m_buf[0]));
				if(m_buf[0]==0)
				{
					DEBUGF(MYSQL_DEBUG,("(OK)\n"));
					cmd->err = MYSQL_OK;
				}
				else
				{
					m_buf[m_len] = 0;
					DEBUGF(MYSQL_DEBUG,("(SQL Error %d) : %s\n", *((uint16_t*)&m_buf[1]), &m_buf[9])); //LE
					cmd->err = MYSQL_SQL;
				}

				netbuf_delete( in );
			}
		}else
			cmd->err = MYSQL_SQL;

//		sys_sem_signal(&cmd->lock);

//	}


}

static struct netbuf *readSock(struct netconn *conn,byte *buf, unsigned short *buflen){

	struct netbuf  *inbuf;
	char *b;

	int err = netconn_recv( conn,&inbuf);

	if(err != ERR_OK){
		m_len = 0;
		return NULL;
	}


	netbuf_data( inbuf, ( void * )&b, buflen )
	/* Process Packet Header */;
	m_len = *((uint16_t*)&b[0]);
	m_packetId = buf[3];
	DEBUGF(MYSQL_DEBUG,("Packet Id %d of length %d\n", b[3], m_len));
	if(m_len > BUF_SIZE)
		memcpy(buf,&b[4],BUF_SIZE);
	else
		memcpy(buf,&b[4],m_len);
	m_packetId++;

	//DEBUGF(LOG,("WARN: Incomplete packet\n"));

	return inbuf;
}

err_t writeSock(struct netconn *conn) //Copy from buf
{
  byte head[4] = { 0 };
  *((uint16_t*)&head[0]) = m_len;
  head[3] = m_packetId;
  DEBUGF(MYSQL_DEBUG,("Packet Id %d\n", head[3]));
  m_packetId++;
  err_t err = netconn_write( conn, head, 4, NETCONN_COPY );
  if(err == ERR_OK)
	err = netconn_write( conn, m_buf, m_len, NETCONN_COPY );
  if(err != ERR_OK)//Error
  {
     return -1;
  }

  return err;
}


int MySqlCommand(struct netconn *conn, byte command, int len){
	byte head[5] = { 0 };
	*((uint16_t*)&head[0]) = len;
	head[3] = m_packetId;
	head[4] = command;
	DEBUGF(MYSQL_DEBUG,("Packet Id %d\n", head[3]));
	m_packetId++;
	err_t err = netconn_write( conn, head, 5, NETCONN_COPY );
	return err;
}

/*
 * A wrapper function, so we know we are sending to mysql
 */
int MySqlSend(struct netconn *conn, byte *data, int len){
	err_t err = netconn_write( conn, data, len, NETCONN_COPY );
	return err;
}

static err_t writeSockCmd(struct netconn *conn, byte command, char *arg)
{
  byte head[5] = { 0 };
  *((uint16_t*)&head[0]) = m_len;
  head[3] = m_packetId;
  head[4] = command;
  DEBUGF(MYSQL_DEBUG,("Packet Id %d\n", head[3]));
  m_packetId++;
  err_t err = netconn_write( conn, head, 5, NETCONN_COPY );
  if(err == ERR_OK)
	  err = netconn_write( conn, arg, m_len, NETCONN_COPY );
  if(err != ERR_OK)//Error
  {
     return -1;
  }

  return err;
}



int MySqlExit(struct netconn *conn){
	  /* close */
   	netconn_close( conn );
   	netconn_delete( conn );
   	return 0;
}

